
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: _G, C_Texture, coroutine, Grid2Frame, Grid2Layout, Grid2LayoutFrame, IsAddOnLoaded, next

local C_IncomingSummon = C_IncomingSummon
local InCombatLockdown = InCombatLockdown
local Mixin = Mixin

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration Grid2")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.numActiveButtons = 0

function M.AtlasToFile(atlas)
	return (C_Texture.GetAtlasInfo(atlas) or {}).file
end

function M.AtlasToTexCoord(atlas)
	local info = C_Texture.GetAtlasInfo(atlas) or {}
	return info.leftTexCoord, info.rightTexCoord, info.topTexCoord, info.bottomTexCoord
end

function M:OnInitialize()
	self.DungeonRoleIcon = self.AtlasToFile("groupfinder-icon-role-large-dps")

	self.DungeonRoleTexCoord = setmetatable({
		DAMAGER = { self.AtlasToTexCoord("groupfinder-icon-role-large-dps") },
		HEALER = { self.AtlasToTexCoord("groupfinder-icon-role-large-heal") },
		TANK = { self.AtlasToTexCoord("groupfinder-icon-role-large-tank") },
	}, { __index = function () return {} end })

	self.StatusIcon = setmetatable({
		["Interface\\RaidFrame\\ReadyCheck-NotReady"] = self.AtlasToFile("ui-lfg-declinemark-raid"),
		["Interface\\RaidFrame\\ReadyCheck-Ready"] = self.AtlasToFile("ui-lfg-readymark-raid"),
		["Interface\\RaidFrame\\ReadyCheck-Waiting"] = self.AtlasToFile("ui-lfg-pendingmark-raid"),
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_1"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/1]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_2"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/2]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_3"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/3]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_4"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/4]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_5"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/5]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_6"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/6]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_7"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/7]],
		["Interface\\TargetingFrame\\UI-RaidTargetingIcon_8"] = [[Interface/AddOns/SixUI/Resources/Textures/RaidTargetIcon/8]],
	}, { __index = function (_, k) return k end })

	self.StatusTexCoord = setmetatable({
		["Interface\\RaidFrame\\ReadyCheck-NotReady"] = { self.AtlasToTexCoord("ui-lfg-declinemark-raid") },
		["Interface\\RaidFrame\\ReadyCheck-Ready"] = { self.AtlasToTexCoord("ui-lfg-readymark-raid") },
		["Interface\\RaidFrame\\ReadyCheck-Waiting"] = { self.AtlasToTexCoord("ui-lfg-pendingmark-raid") },
	}, { __index = function () return {} end })

	self.SummonIcon = self.AtlasToFile("Raid-Icon-SummonPending")

	self.SummonTexCoord = setmetatable({
		[1] = { self.AtlasToTexCoord("Raid-Icon-SummonPending") },
		[2] = { self.AtlasToTexCoord("Raid-Icon-SummonAccepted") },
		[3] = { self.AtlasToTexCoord("Raid-Icon-SummonDeclined") },
	}, { __index = function () return {} end })
end

function M:OnEnable()
	if not IsAddOnLoaded("Grid2") then
		self:RegisterEvent("ADDON_LOADED")
		return
	end

	if not self:IsHooked("CreateFrame") then
		self:SecureHook("CreateFrame")
	end

	self:RegisterMessage("SIXUI_CLUTTER_STATE_CHANGED")
	self:RegisterMessage("SIXUI_CONFIG_UPDATE")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED", "Update")
	self:RegisterMessage("SIXUI_PROFILE_UPDATE", "Update")
	self:HookAllButtons()
	self:Update()
end

function M:OnDisable()
	self:UnhookAllButtons()
	self:RehookStatuses(true)
end

function M:Update()
	self:AdjustButtons()
	self:AdjustLayoutFrame()
	self:RehookStatuses()
end

-- Determine the most relevant raid frame config.
function M:GetRaidFrameConfig()
	return A:GetRaidFrameConfig(self.numActiveButtons)
end

function M:AdjustButtons(tick)
	if self.timer then
		self:CancelTimer(self.timer)
	end

	if not tick then
		self.timer = self:ScheduleTimer("AdjustButtons", 0.1, true)
		return
	end

	if InCombatLockdown() then
		A.CombatQueue:Add(self.AdjustButtons, self.AdjustButtons, self, true)
	else
		local raidFrameConfig = self:GetRaidFrameConfig()
		local profile = Grid2Frame.db.profile

		Mixin(profile, {
			frameWidth = raidFrameConfig.width,
			frameHeight = raidFrameConfig.height,
		})

		Grid2Layout:UpdateFramesSize()
	end
end

function M:AdjustLayoutFrame()
	if InCombatLockdown() then
		A.CombatQueue:Add(self.AdjustLayoutFrame, self.AdjustLayoutFrame, self)
	else
		local raidFrameConfig = self:GetRaidFrameConfig()

		A:RegisterFrame("Grid2LayoutFrame"):ClearAllPoints():SetPoint(raidFrameConfig.point, raidFrameConfig.relativeTo, raidFrameConfig.relativePoint, raidFrameConfig.offsetX, raidFrameConfig.offsetY)
	end
end

function M:DecrementNumActiveButtons()
	self.numActiveButtons = self.numActiveButtons - 1
	self:AdjustButtons()
end

function M:IncrementNumActiveButtons()
	self.numActiveButtons = self.numActiveButtons + 1
	self:AdjustButtons()
end

-- Itereate over Grid2's buttons.
function M:IterateButtons()
	if (not Grid2LayoutFrame) then return end

	return coroutine.wrap(function ()
		for _, header in next, {Grid2LayoutFrame:GetChildren()} do
			if (header.GetName and header:GetName() or ""):match("Grid2LayoutHeader%d+") then
				for _, button in next, {header:GetChildren()} do
					if (button.GetName and button:GetName() or ""):match("Grid2LayoutHeader%d+UnitButton%d+") then
						coroutine.yield(button)
					end
				end
			end
		end
	end)
end

function M:HookAllButtons()
	for button in self:IterateButtons() do
		self:HookButton(button)
	end
end

function M:HookButton(button)
	if (not button) then return end

	if not self:IsHooked(button, "OnShow") then
		self:HookScript(button, "OnShow", "SkinButton")
	end

	if not self:IsHooked(button, "OnHide") then
		self:HookScript(button, "OnHide", "UnskinButton")
	end

	self[button:IsShown() and "SkinButton" or "UnskinButton"](self, button)
end

function M:UnhookAllButtons()
	for button in self:IterateButtons() do
		self:UnhookButton(button)
	end
end

function M:UnhookButton(button)
	self:Unhook(button, "OnHide")
	self:Unhook(button, "OnShow")
	self:UnskinButton(button)
end

function M:SkinButton(button)
	B:CreateBorder(button, C.BORDER_PADDING):SetIgnoreParentAlpha(true)

	if not self:IsHooked(button, "SetBackdropBorderColor") then
		self:RawHook(button, "SetBackdropBorderColor", "OnSetBackdropBorderColor", true)
		self.hooks[button].SetBackdropBorderColor(button, 0, 0, 0, 0)
		self:OnSetBackdropBorderColor(button, button:GetBackdropBorderColor())
		self:IncrementNumActiveButtons()
	end
end

function M:UnskinButton(button)
	B:RemoveBorder(button)

	if self:IsHooked(button, "SetBackdropBorderColor") then
		self:Unhook(button, "SetBackdropBorderColor")
		self:DecrementNumActiveButtons()
	end
end

function M:RehookStatuses(unhook)
	local DungeonRole = Grid2:GetStatusByName("dungeon-role")
	local RaidIconPlayer = Grid2:GetStatusByName("raid-icon-player")
	local RaidIconTarget = Grid2:GetStatusByName("raid-icon-target")
	local ReadyCheck = Grid2:GetStatusByName("ready-check")
	local Summon = Grid2:GetStatusByName("summon")

	local function Rehook(enabled, Status, method, handler)
		if Status then
			if unhook or not enabled then
				self:Unhook(Status, method)
			elseif not self:IsHooked(Status, method) then
				self:RawHook(Status, method, handler)
			end
		end
	end

	Rehook(self.config.useOverrideRoleIcons, DungeonRole, "GetIcon", "DungeonRole_GetIcon")
	Rehook(self.config.useOverrideRoleIcons, DungeonRole, "GetTexCoord", "DungeonRole_GetTexCoord")
	Rehook(self.config.showRoleIconsWithClutterKey, DungeonRole, "IsActive", "DungeonRole_IsActive")
	Rehook(self.config.useOverrideRaidIcons, RaidIconPlayer, "GetIcon", "Status_GetIcon")
	Rehook(self.config.useOverrideRaidIcons, RaidIconTarget, "GetIcon", "Status_GetIcon")
	Rehook(self.config.useOverrideReadyCheckIcons, ReadyCheck, "GetIcon", "Status_GetIcon")
	Rehook(self.config.useOverrideReadyCheckIcons, ReadyCheck, "GetTexCoord", "Status_GetTexCoord")
	Rehook(self.config.useOverrideSummonIcons, Summon, "GetIcon", "Summon_GetIcon")
	Rehook(self.config.useOverrideSummonIcons, Summon, "GetTexCoord", "Summon_GetTexCoord")

	self:UpdateStatuses()
end

function M:UpdateStatuses()
	Grid2Frame:UpdateIndicators()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:ADDON_LOADED(event, name)
	if name == "Grid2" then
		self:UnregisterEvent("ADDON_LOADED")
		self:OnEnable()
	end
end

function M:SIXUI_CLUTTER_STATE_CHANGED(event, state)
	self.isClutterKeyDown = state
	self:UpdateStatuses()
end

function M:SIXUI_CONFIG_UPDATE(event, module, config, key, ...)
	if module == "Core RaidFrameConfig" then
		if key == "width" or key == "height" then
			self:AdjustButtons()
		else
			self:AdjustLayoutFrame()
		end
	elseif module == "Integration Grid2" then
		self:Update()
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:CreateFrame(frameType, frameName, parentFrame, inheritsFrame)
	if frameName then
		if frameName == "Grid2LayoutFrame" then
			self:AdjustLayoutFrame()
		elseif frameName:match("Grid2LayoutHeader%d+UnitButton%d+") then
			self:HookButton(_G[frameName])
		end
	end
end

function M:DungeonRole_GetIcon()
	return self.DungeonRoleIcon
end

function M:DungeonRole_GetTexCoord(_, unit)
	return unpack(self.DungeonRoleTexCoord[UnitGroupRolesAssigned(unit)])
end

function M:DungeonRole_IsActive(DungeonRole, ...)
	return self.isClutterKeyDown and self.hooks[DungeonRole].IsActive(DungeonRole, ...)
end

function M:OnSetBackdropBorderColor(button, r, g, b)
	if button.GetBorder then
		if r ~= 0 or g ~= 0 or b ~= 0 then
			button:GetBorder():SetColor(r, g, b)
		else
			button:GetBorder():SetColor()
		end
	end
end

function M:Status_GetIcon(Status, ...)
	return self.StatusIcon[self.hooks[Status].GetIcon(Status, ...)]
end

function M:Status_GetTexCoord(Status, ...)
	return unpack(self.StatusTexCoord[self.hooks[Status].GetIcon(Status, ...)])
end

function M:Summon_GetIcon()
	return self.SummonIcon
end

function M:Summon_GetTexCoord(_, unit)
	return unpack(self.SummonTexCoord[C_IncomingSummon.IncomingSummonStatus(unit)])
end
